#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#define NFC1 19
#define NFC2 9
#define ODMAX 13
#define STEPMIN 1e-6

double std(double y, double x, double r, double k)
{
 double xr,yr;

  xr=x/r;yr=y/r;
  return ((x*xr+y*yr)/(1+sqrt(1-(1+k)*(xr*xr+yr*yr))));
}

double zstd(double y, double x, double r, double k, double yc, double xc, double zc, double rot, double v[])
{
  double x0,y0,z0,sn,cs,tn,t,xt,yt,zt,vx,vy,vz,vv,a,b,c,t1,t2;

  sn=sin(rot/180*M_PI);
  cs=cos(rot/180*M_PI);
  tn=sn/cs;

  x0=x-xc;
  y0= (y-yc)*cs-zc*sn;
  z0=-(y-yc)*sn-zc*cs;
  xt=x0;
  yt=y0;
  zt=z0;
  a=tn*tn+(1+k);
  b=tn*y0+(1+k)*z0-r;
  c=x0*x0+y0*y0+(1+k)*z0*z0-2*r*z0;
  t1=(-b-sqrt(b*b-a*c))/a;
  t2=(-b+sqrt(b*b-a*c))/a;
  if(fabs(z0+t1) < fabs(z0+t2)) t=t1;
  else t=t2;
  zt+=t;
  yt+=tn*t;
  if(fabs(xt)<1e-9 && fabs(yt)<1e-9){
    v[0]=0;
    v[1]=-sn;
    v[2]=cs;
  }else{
    vx=xt;
    vy=yt;
    vz=(1+k)*zt-r;
    vv=sqrt(vx*vx+vy*vy+vz*vz);
    if(vz<0.){
      vx/=-vv;
      vy/=-vv;
      vz/=-vv;
    }else{
      vx/=vv;
      vy/=vv;
      vz/=vv;
    }
    v[0]=vx;
    v[1]=vy*cs-vz*sn;
    v[2]=vy*sn+vz*cs;
  }
  return ((zt-z0)/cs);
}

double bc(double y, double x, double ry, double rx, double ky, double kx)
{
  double xr,yr;

  xr=x/rx;yr=y/ry;
  return ((x*xr+y*yr)/(1+sqrt(1-(1+kx)*xr*xr-(1+ky)*yr*yr)));
}

double zbc(double y, double x, double ry, double rx, double ky, double kx, double yc, double xc, double zc, double rot, double v[])
{
  double x0,y0,z0,sn,cs,tn,t,xt,yt,zt,xr,yr,xk,yk,vx,vy,vz,vv;

  sn=sin(rot/180*M_PI);
  cs=cos(rot/180*M_PI);
  tn=sn/cs;

  x0=x-xc;
  y0= (y-yc)*cs-zc*sn;
  z0=-(y-yc)*sn-zc*cs;
  xt=x0;
  yt=y0;
  zt=z0;
  t = bc(yt,xt,ry,rx,ky,kx)-zt;
  while(fabs(t)>1e-7){
    yt+=tn*t;
    zt+=t;
    t = bc(yt,xt,ry,rx,ky,kx)-zt;
  }
  zt+=t;
  yt+=tn*t;
  if(fabs(xt)<1e-9 && fabs(yt)<1e-9){
    v[0]=0;
    v[1]=-sn;
    v[2]=cs;
  }else{
    xr=xt*xt/rx;
    yr=yt*yt/ry;
    xk=(1+kx)*zt/rx;
    yk=(1+ky)*zt/ry;
    vx=xt/rx*(2*xr+2*yr+(xk-2)*zt);
    vy=yt/ry*(2*xr+2*yr+(yk-2)*zt);
    vz=(xk-1)*xr+(yk-1)*yr;
    vv=sqrt(vx*vx+vy*vy+vz*vz);
    if(vz<0.){
      vx/=-vv;
      vy/=-vv;
      vz/=-vv;
    }else{
      vx/=vv;
      vy/=vv;
      vz/=vv;
    }
    v[0]=vx;
    v[1]=vy*cs-vz*sn;
    v[2]=vy*sn+vz*cs;
  }
  return ((zt-z0)/cs);
}

double pw(double n, int k)
{
  int kk;
  double nn=1.;
  for(kk=1;kk<=k;kk++) nn*=n;
  return (nn);
}

int kai(int n)
{
  int kk,nn=1;
  for(kk=1;kk<=n;kk++) nn*=kk;
  return (nn);
}

double zernike(int n, int m, double a, double b)
{
  int k;
  double r=0.;

  for(k=0;k<=(n-abs(m))/2;k++) r+=pw(-1.,k)*kai(n-k)*pw(a,n-2*k)/kai(k)/kai((n+m)/2-k)/kai((n-m)/2-k);
  if(m>0) r*=cos(b*m/180.*M_PI);
  else if(m<0) r*=sin(b*(-m)/180.*M_PI);
  return(r);
}

double foccor(double a, double b, double zn[ODMAX][ODMAX])
{
  int m,n;
  double z;

  z=0.;
  a=sin(a/180.*M_PI)*38.3/12/1.05;
  for(n=0;n<ODMAX;n++){
    for(m=-n;m<=n;m+=2) z+=zn[(n+m)/2][(n-m)/2]*zernike(n,m,a,b);
  }
  return (z);
}

double foccor2(double a, double b, double c[NFC1][NFC2])
{
  int i,j;
  double akai[NFC2],fc,z;

  for(j=0;j<NFC2;j++) akai[j]=pw(a,j+2);
  z=0.;
  for(i=0;i<NFC1;i++){
    fc=0.;
    for(j=0;j<NFC2;j++) fc+=c[i][j]*akai[j];
    if(i==0) z=fc;
    else if(i%2) z+=fc*cos(b*((int)(i+1)/2)/180.*M_PI);
    else z+=fc*sin(b*((int)(i+1)/2)/180.*M_PI);
  }
  return (z);
}

double NH3cor(char *fname0, double dx, double dy, int op)
{
  FILE *fp,*fp2,*fpm;
  char buf[1024],fname[32];
  int type,i,j,im,jm,mk,m,n;
  double x,y,z,zm,rx,ry,kx,ky,xc,yc,zc,rot,v[3],dzs,dz2s,dz0,a,b,r,c[NFC1][NFC2],zn[ODMAX][ODMAX],zm1,zm2;

  if((fp = fopen("NH3cor1.dat","r")) == NULL){
    fprintf(stderr,"Cannot open NH3cor1.dat\n");
    exit(1);
  }
  for(n=0;n<ODMAX;n++){
    for(m=0;m<ODMAX-n;m++) fscanf(fp,"%lf",&zn[n][m]);
  }
  fclose(fp);

  if((fp = fopen("NH3cor2.dat","r")) == NULL){
    fprintf(stderr,"Cannot open NH3cor2.dat\n");
    exit(1);
  }
  for(i=0;i<NFC1;i++){
    for(j=NFC2-1;j>=0;j--) fscanf(fp,"%lf",&c[i][j]);
  }
  fclose(fp);

  if((fp = fopen("NH3cor.dat","r")) == NULL){
    fprintf(stderr,"Cannot open NH3cor.dat\n");
    exit(1);
  }
  while(!feof(fp)){
    fgets(buf,sizeof(buf),fp);
    if(buf[0]=='#') continue;
    if(sscanf(buf,"%lf %lf %lf %lf %lf %lf %lf %lf",&ry,&ky,&rx,&kx,&yc,&xc,&zc,&rot)!=8) continue;
  }
  fclose(fp);
  xc+=dx;
  yc+=dy;

  if(fabs(rx-ry)<1e-7 && fabs(kx-ky)<1e-7) type=0;
  else type=1;

  if(op){
    if(type==0){
      fprintf(stderr,"Type     : standard\n");
      fprintf(stderr,"Radius   : %g\n",ry);
      fprintf(stderr,"Conic    : %g\n",ky);
      fprintf(stderr,"Center_y : %g\n",yc);
      fprintf(stderr,"Center_x : %g\n",xc);
      fprintf(stderr,"Center_z : %g\n",zc);
      fprintf(stderr,"Rot_x    : %g\n",rot);
    }
    else{
      fprintf(stderr,"Type     : biconic\n");
      fprintf(stderr,"Radius_y : %g\n",ry);
      fprintf(stderr,"Conic_y  : %g\n",ky);
      fprintf(stderr,"Radius_x : %g\n",rx);
      fprintf(stderr,"Conic_x  : %g\n",kx);
      fprintf(stderr,"Center_y : %g\n",yc);
      fprintf(stderr,"Center_x : %g\n",xc);
      fprintf(stderr,"Center_z : %g\n",zc);
      fprintf(stderr,"Rot_x    : %g\n",rot);
    }
    fprintf(stderr,"Open     : %s\n",fname0);
  }
  if((fp = fopen(fname0,"r")) == NULL){
    fprintf(stderr,"Cannot open %s\n",fname0);
    exit(0);
  }
  if(op){
    sscanf(fname0,"%s",fname);
    strcpy(fname+strlen(fname)-4,"_2.CSV");
    fprintf(stderr,"Save     : %s\n",fname);
    if((fp2 = fopen(fname,"w")) == NULL){
      fprintf(stderr,"Cannot open %s\n",fname);
      exit(0);
    }
  }

  sscanf(fname0,"%s",fname);
  strcpy(fname+strlen(fname)-4,".msk");
  if((fpm = fopen(fname,"r")) == NULL){
    if(op) fprintf(stderr,"Mask     : NONE\n");
    jm=-1;im=-1;
  }
  else{
    if(op) fprintf(stderr,"Mask     : %s\n",fname);
    if(fscanf(fpm,"%d,%d",&jm,&im)!=2){
      fclose(fpm);
      jm=-1;im=-1;
    }
  }

  dz0=0.;
  for(n=0;n<25;n++){
    fgets(buf,sizeof(buf),fp);
    if(op) fprintf(fp2,"%s",buf);
  }
  fgets(buf,sizeof(buf),fp);
  if(op){
    strcpy(buf+strlen(buf)-1,",\"Mask\",\"Z_Model\",\"Beta\",\"Alpha\",\"Residual\",\"Foc Cor1\",\"Res1\",\"Foc Cor2\",\"Res2\"\n");
    fprintf(fp2,"%s",buf);
  }
  n=0;
  dzs=0.;
  dz2s=0.;
  while(!feof(fp)){
    if(fgets(buf,sizeof(buf),fp)==NULL){
      if(op) fprintf(fp2,"\n");
      continue;
    }
    if(buf[0]=='#'){
      if(op) fprintf(fp2,"%s",buf);
      continue;
    }
    if(sscanf(buf,"%d,%d,%lf,%lf,%lf",&j,&i,&x,&y,&z)!=5){
      if(sscanf(buf,"%d,%d,%lf,%lf",&j,&i,&x,&y)!=4 || i!=im || j!=jm){
	if(op) fprintf(fp2,"%s",buf);
	continue;
      }
    }
    if(type==0) zm=zstd(y,x,rx,kx,yc,xc,zc,rot,v);
    else zm=zbc(y,x,ry,rx,ky,kx,yc,xc,zc,rot,v);
    r=sqrt(v[0]*v[0]+v[1]*v[1]);
    a=asin(r)/M_PI*180;
    if(v[1]>=0. && v[1]<=v[0]) b=asin(v[1]/r)/M_PI*180;
    else if(v[1]>v[0] && v[1]>-v[0]) b=acos(v[0]/r)/M_PI*180;
    else if(v[1]<=-v[0] && v[1]>=v[0]) b=-asin(v[1]/r)/M_PI*180+180;
    else if(v[1]<-v[0] && v[1]<v[0]) b=-acos(v[0]/r)/M_PI*180+360;
    else b=asin(v[1]/r)/M_PI*180+360;
    zm1=foccor(a,b,zn);
    zm2=foccor2(a,b,c);
    if(i==im && j==jm){
      mk=1;
      z=zm+(zm1+zm2)/1000+dz0;
      if(fscanf(fpm,"%d,%d",&jm,&im)!=2){
	fclose(fpm);
	jm=-1;im=-1;
      }
    }
    else mk=0;
    if(op){
      fprintf(fp2,"%d,%d,%.5lf,%.5lf,%.5lf,%d,%.5lf,%.5lf,%.5lf,%.5lf,%.5lf,%.5lf,%.5lf,%.5lf\n",j,i,x,y,z,mk,zm,b,a,(z-zm)*1000.,zm1,(z-zm)*1000.-zm1,zm2,(z-zm)*1000.-zm1-zm2);
    }
    //if(x*x+y*y<9.){
      dzs+=(z-zm)*1000.-zm1-zm2;
      dz2s+=((z-zm)*1000.-zm1-zm2)*((z-zm)*1000.-zm1-zm2);
      n++;
      //}
    dz0=z-zm-(zm1+zm2)/1000;
  }
  if(jm!=-1 && im!=-1) fclose(fpm);
  if(op){
    fclose(fp2);
    fprintf(stderr,"Center_z : %g\n",zc+dzs/n);
    fprintf(stderr,"Stddev_dz: %g\n",sqrt(dz2s*n-dzs*dzs)/n);
  }
  fclose(fp);

  return (sqrt(dz2s*n-dzs*dzs)/n);
}

int main(int argc,char *argv[])
{
  double dzsdv,dzsdvmin,dr,dx,dy;
  int i,j,imin,jmin;

  if(argc!=2 && argc != 3){
    fprintf(stderr,"Usage : %s filename [step]\n",argv[0]);
    exit(0);
  }
  if(argc==3){
    sscanf(argv[2],"%lf",&dr);
    dx=0.;
    dy=0.;
    dzsdvmin=1e99;
    while(dr>STEPMIN){
      for(i=-1;i<=1;i++){
	for(j=-1;j<=1;j++){
	  dzsdv=NH3cor(argv[1],dx+dr*j,dy+dr*i,0);
	  if(dzsdvmin>dzsdv){
	    dzsdvmin=dzsdv;
	    imin=i;
	    jmin=j;
	    fprintf(stderr,"%d %d %g %g %g %lf\n",imin,jmin,dr,dy+dr*i,dx+dr*j,dzsdvmin);
	  }
	}
      }
      if(imin==0 && jmin==0){
	dr/=10.;
	dzsdvmin+=1e-10;
	continue;
      }
      else{
	dx+=dr*jmin;
	dy+=dr*imin;
	while(1){
	  dx+=dr*jmin;
	  dy+=dr*imin;
	  dzsdv=NH3cor(argv[1],dx+dr*j,dy+dr*i,0);
	  if(dzsdvmin>dzsdv){
	    fprintf(stderr,"%d %d %g %g %g %lf\n",imin,jmin,dr,dy,dx,dzsdvmin);
	    dzsdvmin=dzsdv;
	  }
	  else{
	    dx-=dr*jmin;
	    dy-=dr*imin;
	    dzsdvmin+=1e-10;
	    break;
	  }
	}
      }
    }
    
    fprintf(stderr,"Center_dy: %g\n",dy);
    fprintf(stderr,"Center_dx: %g\n",dx);
    fprintf(stderr,"Stddev_dz: %g\n",dzsdvmin);
  }
  else NH3cor(argv[1],0.,0.,1);
  return 0;
}
